/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Install
Version:   1.1
Status:    Stable
ID:        ID_INSTALL
Category:  Tool
Date:      July 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2002-2004.  All rights reserved.
Keywords:  install, installation, setup
Short:     Installs programs and other forms of software and data.

###DESCRIPTION###
<p>The install class is provided for the purpose of setting a standard in the
installation of Athene and DML based software.  There is only one way to use
this class, which involves building an application interface around it.
The QikInstall program, which is distributed with Athene and DML shows how to
correctly implement the Install class in an application.  We recommend that
you refer to the QikInstall DML scripts to see just how this is achieved.</p>

<p>If you would like information on the XML-Install standard or need to know
how install packages are created, please refer to the Pandora SDK Manual.  The
section entitled &quot;Distribution and Installation&quot; will give you all
the information that you are looking for.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/files/all.h>
#include <pandora/data/config.h>
#include <pandora/data/compression.h>
#include <pandora/data/xml.h>
#include <pandora/graphics/render.h>
#include <pandora/graphics/scroll.h>
#include <pandora/graphics/text.h>
#include <pandora/input/onclick.h>
#include <pandora/misc/time.h>
#include <pandora/tools/filedialog.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/functions/runscript.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_INSTALL 1.2

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_INSTALL, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2002-2004.  All rights reserved.",
   "July 2004",
   "Install"
};

static struct KernelBase  *KernelBase = 0;
static struct StringsBase *StringsBase = 0;
static struct FileSystemBase *FileSystemBase = 0;
static OBJECTPTR InstallClass = 0;
static OBJECTPTR modStrings = 0;
static OBJECTPTR modFileSystem = 0;
static FIELD FID_OkayText;

#define LEN_FONTFACE 80

static UBYTE glFontFace[LEN_FONTFACE] = "[gl_fonts.face]";

/*****************************************************************************
** Class definition.
*/

#define MAX_VARIABLES 40
#define MAX_DIRECTORIES 10
#define MAX_OPTIONS 40

struct InstallVariable {
   STRING Name;
   STRING Value;
};

struct InstallDirectory {
   OBJECTPTR Text;
   STRING    Name;
   STRING    Path;
};

struct InstallOption {
   OBJECTID ButtonID;
   STRING   Name;
   WORD     OptionStatus;
   WORD     Level;
   WORD     XCoord;
   WORD     YCoord;
};

struct Install {
   OBJECT_HEADER            /* Class header */
   OBJECTID NavigationID;   /* Reference to the file navigation area */
   OBJECTID ViewID;         /* Reference to the viewing area */
   OBJECTID VSViewID;       /* The vertical scrollbar assigned to the view */
   OBJECTID VSNavigationID; /* The vertical scrollbar assigned to the navigator */
   OBJECTID HSViewID;       /* The horizontal scrollbar assigned to the view */
   OBJECTID HSNavigationID; /* The horizontal scrollbar assigned to the navigator */
   OBJECTID CancelButtonID; /* Cancel button ID */
   OBJECTID NextButtonID;   /* Next button ID */
   OBJECTID PrevButtonID;   /* Previous button ID */
   LONG Test;               /* If TRUE, enables test-mode, which prevents operations during installation */

   PRIVATE_FIELDS
   struct InstallVariable  prvVariables[MAX_VARIABLES];
   struct InstallDirectory prvDirectories[MAX_DIRECTORIES];
   struct InstallOption    prvOptions[MAX_OPTIONS];
   struct XMLTag **prvTags;
   struct XMLTag *prvTag;
   LONG prvInstallIndex;
   OBJECTPTR prvCompression;
   OBJECTPTR prvXML;
   OBJECTPTR prvConfig;
   OBJECTPTR prvUninstall;
   OBJECTPTR prvOutput;
   STRING prvLocation;
   STRING prvMergeSection;
   STRING prvPackageName;
   LONG prvTotalVariables;
   LONG prvDirectoryCount;
   LONG prvOptionCount;
   LONG prvCurrentXCoord;
   LONG prvCurrentYCoord;
   LONG prvTotalTags;
   LONG prvIndent;
   LONG prvIndentLevel;
   LONG prvBoxSize;
   LONG prvUninstallIndex;
   LARGE prvSysTime;
   WORD prvCurrentTagID;
   WORD prvInstalled;
   WORD prvRefreshDesktop;
   FLOAT prvKernelVersion;
   FLOAT prvInstallVersion;
   BYTE NoSUIDOption;   /* For copy operations */
   BYTE Permanent;      /* For copy operations */
   LONG Permissions;    /* For copy operations */
   BYTE NoFail;         /* Incrementing this value stops the install process from cancelling due to failure */
   BYTE Quiet;
   BYTE Processing;     /* The installation commands are currently being procesed */
   BYTE CancelRequest;  /* Set to TRUE to indicate that a request has been cancelled */
};

static ERROR GET_Location(struct Install *, STRING *);

static ERROR SET_Location(struct Install *, STRING);

static struct FieldArray InstallFields[] = {
   { "Navigation",   0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "View",         0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "VSView",       0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "VSNavigation", 0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "HSView",       0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "HSNavigation", 0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "CancelButton", 0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "NextButton",   0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "PrevButton",   0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   { "Test",         0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   /*** Virtual fields ***/
   { "Location", 0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   { "Src",      0, FDF_STRING|FDF_RW, 0, GET_Location, SET_Location },
   END_FIELD
};

static ERROR INSTALL_ActionNotify(struct Install *, struct acActionNotify *);
static ERROR INSTALL_Activate(struct Install *, APTR);
static ERROR INSTALL_Free(struct Install *, APTR);
static ERROR INSTALL_Init(struct Install *, APTR);

static struct ActionArray InstallActions[] = {
   { AC_Activate,     INSTALL_Activate },
   { AC_ActionNotify, INSTALL_ActionNotify },
   { AC_Free,         INSTALL_Free },
   { AC_Init,         INSTALL_Init },
   { NULL, NULL }
};

/*****************************************************************************
** Method codes for copy operations.
*/

#define METHOD_CAUTIOUS  0
#define METHOD_OVERWRITE 1
#define METHOD_BACKUP    2

/*****************************************************************************
** Function prototypes.
*/

static ERROR AddVariable(struct Install *, STRING, STRING);
static void CancelInstall(struct Install *, WORD);
static LONG CopyFeedback(OBJECTPTR, struct CompressionFeedback *);
static ERROR CreateOption(struct Install *, OBJECTID *, STRING, LONG);
static OBJECTPTR DisplayXML(struct Install *, STRING);
static ERROR ExecuteInstallTags(struct Install *, OBJECTPTR, struct XMLTag *);
static void FinishInstall(struct Install *);
static struct XMLTag * FindTag(struct XMLTag *, STRING);
static void KillChildren(struct Install *);
static struct XMLTag * PreviousTag(struct XMLTag *);
static ERROR ProcessTag(struct Install *, WORD);
static void ResetPage(struct Install *);
static ERROR RunTags(struct Install *);
static void TerminateInstall(struct Install *);

static ERROR CommandComponent(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandCopy(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandDelete(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandDirectory(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandItem(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandLink(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandMerge(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandMove(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandNoFail(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandOption(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandRegister(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandRename(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandRun(struct Install *, OBJECTPTR, struct XMLTag *);
static ERROR CommandSection(struct Install *, OBJECTPTR, struct XMLTag *);

static LONG ProcessBreak(struct Install *, struct XMLTag *);
static LONG ProcessCheck(struct Install *, struct XMLTag *);
static LONG ProcessComponents(struct Install *, struct XMLTag *);
static LONG ProcessDirectories(struct Install *, struct XMLTag *);
static LONG ProcessImage(struct Install *, struct XMLTag *);
static LONG ProcessInfo(struct Install *, struct XMLTag *);
static LONG ProcessInstallation(struct Install *, struct XMLTag *);
static LONG ProcessText(struct Install *, struct XMLTag *);

/****************************************************************************/

#define TAGID_COMPONENTS   1
#define TAGID_DIRECTORIES  2
#define TAGID_IMAGE        3
#define TAGID_INFO         4
#define TAGID_INSTALLATION 5
#define TAGID_TEXT         6
#define TAGID_CHECK        7
#define TAGID_BREAK        8

struct InstallTag {
   LONG TagID;
   STRING Tag;
   LONG   (*Routine)(struct Install *, struct XMLTag *);
};

struct InstallCommand {
   STRING Tag;
   ERROR (*Routine)(struct Install *, OBJECTPTR, struct XMLTag *);
};

struct InstallTag glTags[] = {
   { TAGID_BREAK,        "break",        ProcessBreak },
   { TAGID_CHECK,        "check",        ProcessCheck },
   { TAGID_COMPONENTS,   "components",   ProcessComponents },
   { TAGID_DIRECTORIES,  "directories",  ProcessDirectories },
   { TAGID_IMAGE,        "image",        ProcessImage },
   { TAGID_INFO,         "info",         ProcessInfo },
   { TAGID_INSTALLATION, "installation", ProcessInstallation },
   { TAGID_TEXT,         "text",         ProcessText },
   { NULL, NULL }
};

struct InstallCommand glInstallCommands[] = {
   { "component",   CommandComponent },
   { "copy",        CommandCopy },
   { "delete",      CommandDelete },
   { "directory",   CommandDirectory },
   { "merge",       CommandMerge },
   { "item",        CommandItem }, /* child of the merge command */
   { "section",     CommandSection }, /* child of the merge command */
   { "move",        CommandMove },
   { "option",      CommandOption },
   { "register",    CommandRegister },
   { "rename",      CommandRename },
   { "run",         CommandRun },
   { "nofail",      CommandNoFail },
   { "test",        CommandComponent }, /* <test> is a synonym for <component> */
   { "destination", CommandDirectory }, /* deprecated */
   { "link",        CommandLink },
   { NULL, NULL }
};

static const UBYTE txtClickNext[] = "<item/><item><font colour=\"#202080\">Please click the Next button to continue.</font></item>";

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   ResolveFields("OkayText", &FID_OkayText, TAGEND);

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING, "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   StrTranslate(glFontFace, LEN_FONTFACE, NULL);

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&InstallClass, NULL,
      FID_BaseClassID|TLONG, ID_INSTALL,
      FID_Version|TFLOAT,    VER_INSTALL,
      FID_Name|TSTRING,      "Install",
      FID_Category|TLONG,    CCF_TOOL,
      FID_Actions|TPTR,      InstallActions,
      FID_Fields|TPTR,       InstallFields,
      FID_Size|TLONG,        sizeof(struct Install),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (modStrings)    { acFree(modStrings);    modStrings = NULL; }
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = NULL; }
   if (InstallClass)  { acFree(InstallClass);  InstallClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Action: ActionNotify
*/

static ERROR INSTALL_ActionNotify(struct Install *Self, struct acActionNotify *Args)
{
   LONG i, j, frame;
   ERROR error;

   if (!Self->prvTag) {
      if (Self->prvXML) FinishInstall(Self);
      return(ERR_Okay);
   }

   if (Args->ObjectID IS Self->NextButtonID) {

     if (Self->Processing) return(ERR_Okay);

      /*** Go to the next tag ***/

      if (Self->prvCurrentTagID IS TAGID_COMPONENTS) {
         OBJECTPTR button;

         /*** Record the option values ***/

         for (i=0; i < Self->prvOptionCount; i++) {
            if (AccessObject(Self->prvOptions[i].ButtonID, 5000, &button) IS ERR_Okay) {
               if (GetField(button, FID_Frame, FT_LONG, &frame) IS ERR_Okay) {
                  if (frame IS 2) AddVariable(Self, Self->prvOptions[i].Name, "1");
                  else AddVariable(Self, Self->prvOptions[i].Name, "0");
               }
               ReleaseObject(button);
            }
         }
      }
      else if (Self->prvCurrentTagID IS TAGID_DIRECTORIES) {
         STRING path;

         /*** Record the values that the user entered in the text boxes ***/

         for (i=0; i < Self->prvDirectoryCount; i++) {
            if (GetField(Self->prvDirectories[i].Text, FID_String, FT_POINTER, &path) IS ERR_Okay) {
               /*** Check that the path contains an assignment ***/
               if (path) {
                  DPrintF("Installer:","[Install] Path for \"%s\" is \"%s\".", Self->prvDirectories[i].Name, path);
                  for (j=0; (path[j]) AND (path[j] != ':'); j++);
                  if (path[j] IS ':') AddVariable(Self, Self->prvDirectories[i].Name, path);
               }
               else DPrintF("@Installer:","[Install] The directory string for \"%s\" was cleared by the user.", Self->prvDirectories[i].Name);
            }
            else DPrintF("@Installer:","[Install] Failed to retrieve directory string for \"%s\".", Self->prvDirectories[i].Name);
         }
      }
      else if (Self->prvCurrentTagID IS TAGID_INSTALLATION) {
         /*** Process the XML tags inside the <installation> ... </installation> area ***/

         Self->CancelRequest = FALSE;

         acDisableID(Self->PrevButtonID);
         acDisableID(Self->NextButtonID);

         Self->Processing = TRUE;
         error = ExecuteInstallTags(Self, Self->prvOutput, Self->prvTag->Child);
         Self->Processing = FALSE;

         if (Self->CancelRequest) {
            CancelInstall(Self, TRUE);
         }
         else {
            if (error) {
               UBYTE buffer[300];
               STRING *errors;

               if (error IS ERR_NoPermission) {
                  i = StrFormat(buffer, sizeof(buffer), "<item/><item><font colour=\"#a02020\">A permissions failure occurred during the installation process.  You should login as an administrator or super user to install this program.  Install cancelled.</font></item>", error);
                  acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_XML, 1, buffer, i, 1);
               }
               else {
                  acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_XML, 1, "<item/>", 7, 1);

                  if ((errors = (STRING *)(LONG)GetResource(RES_ERRORCODES))) {
                     if (error < GetResource(RES_ERRORCODES_TOTAL)) {
                        if (errors[error]) {
                           i = StrFormat(buffer, sizeof(buffer), "<item><font colour=\"#a02020\">Error %d: \"%s\"</font></item>", error, errors[error]);
                           acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_XML, 1, buffer, i, 1);
                        }
                     }
                  }

                  i = StrFormat(buffer, sizeof(buffer), "<item><font colour=\"#a02020\">An error occurred during the installation process.  Install cancelled.</font></item>");
                  acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_XML, 1, buffer, i, 1);
               }
            }

            if (Self->prvTag) {
               acDataChannel(Self->prvOutput, Self->Head.UniqueID, DATA_XML, 1, txtClickNext, sizeof(txtClickNext), 1);
               acEnableID(Self->NextButtonID);
            }
         }

         /* Scroll to the end of the text list so that the user gets a
         ** visual notifcation that the process is complete.
         */

         struct acScroll scroll;
         scroll.XChange = 0;
         scroll.YChange = 1000000;
         scroll.ZChange = 0;
         ActionMsg(AC_Scroll, Self->ViewID, &scroll);

         acDisableID(Self->CancelButtonID);

         Self->prvInstalled = TRUE;
         Self->prvCurrentTagID = NULL;
         if (Self->prvTag) Self->prvTag = Self->prvTag->Next;
         return(ERR_Okay);
      }

      /*** Process the next tag ***/

      if (Self->prvTag) Self->prvTag = Self->prvTag->Next;
      if (!Self->prvTag) FinishInstall(Self);
      else ProcessTag(Self, 1);
   }
   else if (Args->ObjectID IS Self->PrevButtonID) {
      /*** Cancel the current form's content and move back to the previous tag ***/

      if (Self->Processing) return(ERR_Okay);

      Self->prvTag = PreviousTag(Self->prvTag);
      ProcessTag(Self, -1);
   }
   else if (Args->ObjectID IS Self->CancelButtonID) {
      /* Cancel the installation.  Note that we can only do this if
      ** the user has not reached the <installation> tag yet.
      */

      DPrintF("Install:","User has pressed the cancel button.");

      if (Self->prvInstalled IS FALSE) {
         Self->prvTag = NULL;
         CancelInstall(Self, TRUE);
      }
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: ProcessTag()
*/

ERROR ProcessTag(struct Install *Self, WORD Direction)
{
   LONG i;
   ERROR result;

   DPrintF("~ProcessTag()","[Install]");

   KillChildren(Self);

   /* Note that 'Self->prvTag' is a variable that reflects our current
   ** tag position in the installation process.
   */

   while (Self->prvTag) {
      /*** Find the subroutine that handles the current tag and execute it ***/

      for (i=0; glTags[i].Tag; i++) {
         if (StrCompare(glTags[i].Tag, Self->prvTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            Self->prvCurrentTagID = glTags[i].TagID;
            result = glTags[i].Routine(Self, Self->prvTag);

            if (result IS ERR_False) {
               TerminateInstall(Self);
               StepBack();
               return(ERR_Okay);
            }
            else if (result IS ERR_NoData) {
               KillChildren(Self);
               if (Direction IS -1) Self->prvTag = Self->prvTag->Prev;
               else Self->prvTag = Self->prvTag->Next;
            }
            else if (result != ERR_Okay) {
                DisplayXML(Self, "<item><b><font colour=\"#202080\">Install Error</font></b></item><item/><item>A system error has occurred during installation, causing the install to be cancelled.  You may quit the installer or try the install process again if you wish.</item>");
                TerminateInstall(Self);
                StepBack();
                return(ERR_Okay);
            }
            else {
               /*** Reconfigure the buttons ***/

               if ((Self->prvTag) AND (Self->prvInstalled IS FALSE)) {
                  if (PreviousTag(Self->prvTag) != Self->prvTag) acEnableID(Self->PrevButtonID);
                  else acDisableID(Self->PrevButtonID);
               }

               StepBack();
               return(ERR_Okay);
            }

            break;
         }
      }

      /*** If the tag could not be identified, skip it ***/

      if (!glTags[i].Tag) {
         DPrintF("@Installer:","[Install] Skipping unsupported tag <%s>.", Self->prvTag->Attributes[0].Name);

         if (Direction IS -1) Self->prvTag = Self->prvTag->Prev;
         else Self->prvTag = Self->prvTag->Next;
      }
   }

   StepBack();
   return(ERR_Okay);
}

static ERROR RunTags(struct Install *Self)
{
   WORD i;

   DPrintF("~RunTags()","[Install]");

   /* Note that 'Self->prvTag' is a variable that reflects our current
   ** tag position in the installation process.
   */

   while (Self->prvTag) {
      /*** Find the subroutine that handles the current tag and execute it ***/

      for (i=0; glTags[i].Tag; i++) {
         if (StrCompare(glTags[i].Tag, Self->prvTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            Self->prvCurrentTagID = glTags[i].TagID;
            glTags[i].Routine(Self, Self->prvTag);
            break;
         }
      }

      if (Self->prvTag) Self->prvTag = Self->prvTag->Next;
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Starts the installation process.

###DESCRIPTION###

###RESULT###

###END###

*****************************************************************************/

UBYTE strUninstall[] = "<?xml version=\"1.0\"?>\n\n<uninstall>\n";

static ERROR INSTALL_Activate(struct Install *Self, APTR Void)
{
   struct acWrite write;
   struct mtGetXMLString getxml;
   struct XMLTag **xmllist;
   STRING platform;
   OBJECTPTR file;
   LONG xmltotal, count, i;
   UBYTE start;

   if (!Self->prvLocation) return(ObjectError(Self, ERH_Activate, ERR_MissingLocation));

   DPrintF("~Activate()","[Install:%d] %s", Self->Head.UniqueID, Self->prvLocation);

   /*** Destroy any existing variables from previous activations ***/

   for (i=0; i < Self->prvTotalVariables; i++) {
      FreeMemory(Self->prvVariables[i].Name);
      Self->prvVariables[i].Name  = NULL;
      Self->prvVariables[i].Value = NULL;
   }

   Self->prvTotalVariables = 0;

   /* Extract the install.xml file from the archive and
   ** load it into an XML object.
   */

   if (Self->prvXML) {
      acFree(Self->prvXML);
      Self->prvXML = NULL;
   }

   if (Self->prvCompression) {
      acFree(Self->prvCompression);
      Self->prvCompression = NULL;
   }

   /*** Open a compression object to gain access to the install file ***/

   if (CreateObject(ID_COMPRESSION, NF_CHILD, &Self->prvCompression, NULL,
      FID_Location|TSTRING,  Self->prvLocation,
      FID_Permissions|TLONG, PERMIT_ALL_READ|PERMIT_WRITE|PERMIT_GROUP_WRITE,
      FID_Flags|TLONG,       CMF_READONLY,
      TAGEND) != ERR_Okay) return(ObjectError(Self, ERH_Init, ERR_CreateObject));

   if (ActionTags(MT_DecompressFile, Self->prvCompression, "install.xml", "temp:") IS ERR_Okay) {
      if (CreateObject(ID_XML, NF_CHILD, &Self->prvXML, NULL,
         FID_Location|TSTRING, "temp:install.xml",
         TAGEND) IS ERR_Okay) {

         /*** Delete the decompressed install.xml file ***/

         if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
            FID_Location|TSTRING, "temp:install.xml",
            TAGEND) IS ERR_Okay) {
            Action(MT_DeleteFile, file, NULL);
            acFree(file);
         }
      }
      else {
         DPrintF("@Init:","[Install] Failed to create an XML object.");
         StepBack();
         return(ERR_CreateObject);
      }
   }
   else {
      DPrintF("@Init:","[Install] Failed to locate \"install.xml\" in the install file.");
      StepBack();
      return(ERR_BadData);
   }

   if (GetFields(Self->prvXML, FID_Tags|TPTR,      &xmllist,
                               FID_TagCount|TLONG, &xmltotal,
                               TAGEND) != ERR_Okay) {
      acFree(Self->prvXML);
      Self->prvXML = NULL;
      StepBack();
      return(ObjectError(Self, ERH_Activate, ERR_GetField));
   }

   /*** Check for an <uninstall> tag ***/

   if (Self->Test IS FALSE) {
      Self->prvUninstallIndex = 0;
      for (count=0; xmllist[count]; count++) {
         if (StrCompare("uninstall", xmllist[count]->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            start = FALSE;
            for (i=1; i < xmllist[count]->TotalAttributes; i++) {
               if (StrCompare("start", xmllist[count]->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) start = TRUE;
            }

            /*** Create a temporary uninstall file for this package ***/

            if (Self->prvUninstall) {
               acFree(Self->prvUninstall);
               Self->prvUninstall = NULL;
            }

            if (CreateObject(ID_FILE, NF_CHILD, &Self->prvUninstall, NULL,
               FID_Location|TSTRING, "temp:uninstall.xml",
               FID_Flags|TLONG,      FL_NEW|FL_WRITE,
               TAGEND) IS ERR_Okay) {

               /*** Write the file header ***/

               write.Buffer = strUninstall;
               write.Length = sizeof(strUninstall) - 1;
               if (Action(AC_Write, Self->prvUninstall, &write) != ERR_Okay) {
                  StepBack();
                  return(ObjectError(Self, ERH_Activate, ERR_Write));
               }

               /* If the developer has listed a series of custom commands
               ** in the <uninstall> area then we need to output them (the
               ** default is to place them at the end of the automated
               ** commands, but the developer can override this if the 'start'
               ** instruction is used in the uninstall tag.
               */

               if (xmllist[count]->Child) {
                  if (start) {
                     DPrintF("Install:","Writing custom XML uninstall commands.");
                     getxml.Index = xmllist[count]->Child->Index;
                     getxml.Flags = XMF_INCLUDESIBLINGS;
                     if (Action(MT_GetXMLString, Self->prvXML, &getxml) IS ERR_Okay) {
                        for (i=0; getxml.Result[i]; i++);
                        acWrite(Self->prvUninstall, getxml.Result, i);
                        FreeMemory(getxml.Result);
                     }
                  }
                  else {
                     DPrintF("Install:","Storing custom XML uninstall commands.");
                     Self->prvUninstallIndex = xmllist[count]->Child->Index;
                  }
               }
            }
            else {
               StepBack();
               return(ObjectError(Self, ERH_Activate, ERR_CreateObject));
            }

            break;
         }
      }
   }

   /*** Find the <install> tag ***/

   for (count=0; xmllist[count]; count++) {
      if (StrCompare("install", xmllist[count]->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) break;
   }

   if (!xmllist[count]) {
      DPrintF("!Activate:","[Install] No <install> tag was found in the install.xml file.");
      Action(AC_Free, Self->prvXML, NULL);
      Self->prvXML = NULL;
      StepBack();
      return(ERR_BadData);
   }

   Self->prvTags         = xmllist;
   Self->prvInstallIndex = count;
   Self->prvTotalTags    = xmltotal - count;
   Self->prvInstalled    = FALSE;
   Self->prvPackageName  = NULL;
   Self->prvMergeSection = NULL;
   Self->prvTag          = xmllist[Self->prvInstallIndex]->Child;
   Self->prvRefreshDesktop = FALSE;
   Self->CancelRequest   = FALSE;
   Self->Processing      = FALSE;

   /*** Setup the buttons and navigation area ***/

   DPrintF("Activate:","Setting up buttons.");

   if (Self->NavigationID) acDisableID(Self->NavigationID);
   if (Self->PrevButtonID) acDisableID(Self->NextButtonID);
   if (Self->NextButtonID) acEnableID(Self->NextButtonID);
   if (Self->CancelButtonID) acEnableID(Self->CancelButtonID);

   /*** Add default variables **/

   DPrintF("Activate:","Adding default variables.");

   platform = (STRING)(LONG)GetResource(RES_PLATFORM);
   if (StrCompare("Windows", platform, 0, STR_MATCHLENGTH) IS ERR_Okay) AddVariable(Self, "platform_windows", "1");
   else if (StrCompare("Linux", platform, 0, STR_MATCHLENGTH) IS ERR_Okay) AddVariable(Self, "platform_linux", "1");
   else if ((StrCompare("Athene", platform, 0, STR_MATCHLENGTH) IS ERR_Okay) OR
            (StrCompare("Athenyx", platform, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
      AddVariable(Self, "platform_athene", "1");
      AddVariable(Self, "platform_athenyx", "1");
      AddVariable(Self, "platform_linux", "1");
   }

   /* Start the installation process by calling ProcessTag().  This will run the
   ** first tag and then return immediately.  The rest of the installation process
   ** will be governed by the Next and Prev buttons.
   */

   DPrintF("Activate:","Starting installation process.");

   ProcessTag(Self, 1);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Install: Free
*/

static ERROR INSTALL_Free(struct Install *Self, APTR Void)
{
   OBJECTPTR object;
   LONG i;

   for (i=0; i < Self->prvTotalVariables; i++) {
      FreeMemory(Self->prvVariables[i].Name);
      Self->prvVariables[i].Name  = NULL;
      Self->prvVariables[i].Value = NULL;
   }

   if (Self->prvLocation) {
      FreeMemory(Self->prvLocation);
      Self->prvLocation = NULL;
   }

   if (Self->prvPackageName) {
      FreeMemory(Self->prvPackageName);
      Self->prvPackageName = NULL;
   }

   if (Self->prvUninstall) {
      acFree(Self->prvUninstall);
      Self->prvUninstall = NULL;
   }

   if (Self->prvCompression) {
      acFree(Self->prvCompression);
      Self->prvCompression = 0;
   }

   if (Self->prvXML) {
      acFree(Self->prvXML);
      Self->prvXML = 0;
   }

   if (Self->NextButtonID) {
      if (AccessObject(Self->NextButtonID, 5000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ReleaseObject(object);
      }
   }

   if (Self->PrevButtonID) {
      if (AccessObject(Self->PrevButtonID, 5000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ReleaseObject(object);
      }
   }

   if (Self->CancelButtonID) {
      if (AccessObject(Self->CancelButtonID, 5000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ReleaseObject(object);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Install: Init
*/

static ERROR INSTALL_Init(struct Install *Self, APTR Void)
{
   OBJECTPTR object;
   LONG flags;

   /*** Object validation ***/

   if (!Self->ViewID) {
      DPrintF("@Init:","[Install] Missing reference to a drawable area for viewing.");
      return(ERR_FieldNotSet);
   }

   /*** Set the SCROLLCONTENT flag so that we can scroll the installation render page ***/

   if (AccessObject(GetOwnerID(Self->ViewID), 5000, &object) IS ERR_Okay) {
      GetFields(object, FID_Flags|TLONG, &flags, TAGEND);
      SetFields(object, FID_Flags|TLONG, flags|RNF_SCROLLCONTENT, TAGEND);
      ReleaseObject(object);
   }

   /* Ensure that the page display is of fixed height and not relative
   ** to the window size.
   */
/*
   if (AccessObject(Self->ViewID, 5000, &object) IS ERR_Okay) {
      SetFields(object, FID_Height|TLONG, 3000, TAGEND);
      ReleaseObject(object);
   }
*/
   /*** Subscribe to the user click actions of the cancel, next and prev buttons ***/

   if (Self->NextButtonID) {
      if (AccessObject(Self->NextButtonID, 5000, &object) IS ERR_Okay) {
         SubscribeAction(object, AC_UserClick, Self);
         ReleaseObject(object);
      }
   }

   if (Self->PrevButtonID) {
      if (AccessObject(Self->PrevButtonID, 5000, &object) IS ERR_Okay) {
         SubscribeAction(object, AC_UserClick, Self);
         ReleaseObject(object);
      }
   }

   if (Self->CancelButtonID) {
      if (AccessObject(Self->CancelButtonID, 5000, &object) IS ERR_Okay) {
         SubscribeAction(object, AC_UserClick, Self);
         ReleaseObject(object);
      }
   }

   if (Self->prvLocation) acActivate(Self);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:    Location
Synonym: Src
Short:   The location of the compressed install file is specified here.
Status:  Get/Set
Type:    STRING

###DESCRIPTION###
<p>Before intiailising an install object you will need to set the location
of the package that is to be installed.  The package must have a '.install'
extension and be compressed in a format supported by the
<a href="compression.html">Compression</a> class (pkzip is currently
recommended).  If the file is not valid then the install object will fail
when the Activate action is called.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct Install *Self, STRING *Value)
{
   *Value = Self->prvLocation;
   return(ERR_Okay);
}

static ERROR SET_Location(struct Install *Self, STRING Value)
{
   LONG i;

   if (Self->prvLocation) { FreeMemory(Self->prvLocation); Self->prvLocation = NULL; }

   if ((Value) AND (*Value)) {
      for (i=0; Value[i]; i++);
      if (AllocMemory(i+1, MEM_STRING|MEM_NOCLEAR, (void **)&Self->prvLocation, NULL) IS ERR_Okay) {
         for (i=0; Value[i]; i++) Self->prvLocation[i] = Value[i];
         Self->prvLocation[i] = 0;
      }
      else return(ObjectError(Self, ERH_SetField, ERR_AllocMemory));
   }
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:    CancelButton
Short:   Must refer to a drawable object that will act as a cancel button.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>The CancelButton field must refer to a drawable object that is acting as
a cancel button for the install application.  The inclusion of this button will
allow the user to cancel the install process if he or she wants to terminate
it at any stage.  The following is an example of a cancel button created in
DML:</p>

<pre>
  &lt;button name="btnCancel{id}" xoffset="[win{id}.rightmargin]"
    yoffset="[win{id}.bottommargin]" text="Cancel" disable/&gt;
</pre>

<p>It is recommended that you set the button to disabled mode to start off
with.  The install object will automatically enable and disable the button as
required.</p>

###SEE ALSO###
Fields: NextButton, PrevButton

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    HSNavigation
Short:   Refers to the horizontal scrollbar used for the file navigation area.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>This field needs to be set to the ID of the horizontal scrollbar that has
been created in the file navigation area.  If you have not created a
relevant horizontal scrollbar then there is no need to set this field.</p>

###SEE ALSO###
Fields: HSView, VSNavigation, VSView

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    HSView
Short:   Refers to the horizontal scrollbar used for the viewing area.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>This field needs to be set to the ID of the horizontal scrollbar that has
been created in the viewing area.  If you have not created a relevant
horizontal scrollbar then there is no need to set this field.</p>

###SEE ALSO###
Fields: HSNavigation, VSNavigation, VSView

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    Navigation
Short:   Refers to a drawable that provides the file navigation area.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>If your install application has a file navigation area, provide a link to
its drawable in this field.  This will allow the installer to disable the
navigator during the installation process and even use it for other
purposes.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    NextButton
Short:   Must refer to a drawable object that will act as a next button.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>The NextButton field must refer to a drawable object that is acting as
a next button for the install application.  The inclusion of this button will
allow the user to move between the various stages of the installation process.
The following is an example of a next button created in DML:</p>

<pre>
  &lt;button name="btnNext{id}" xoffset="[win{id}.rightmargin]"
    yoffset="[win{id}.bottommargin]" text="Next >" disable/&gt;
</pre>

<p>It is recommended that you set the button to disabled mode to start off
with.  The install object will automatically enable and disable the button as
required.</p>

###SEE ALSO###
Fields: CancelButton, PrevButton

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    PrevButton
Short:   Must refer to a drawable object that will act as a previous button.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>The PrevButton field must refer to a drawable object that is acting as
a previous button for the install application.  The inclusion of this button will
allow the user to move between the various stages of the installation process.
The following is an example of a previous button created in DML:</p>

<pre>
  &lt;button name="btnPrev{id}" xoffset="[win{id}.rightmargin]"
    yoffset="[win{id}.bottommargin]" text="Previous <" disable/&gt;
</pre>

<p>It is recommended that you set the button to disabled mode to start off
with.  The install object will automatically enable and disable the button as
required.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    View
Short:   Reference to a drawable that provides the wiewing area.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>Your install application must provide a viewing area to which the install
object can draw its text and graphics.  You need to set the View field so that
it refers to the drawable of the viewing area, or the initialisation process
will fail.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    VSNavigation
Short:   Refers to the vertical scrollbar used for the viewing area.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>This field needs to be set to the ID of the vertical scrollbar that has
been created in the file navigation area.  If you have not created a
relevant vertical scrollbar then there is no need to set this field.</p>

###SEE ALSO###
Fields: HSNavigation, HSView, VSView

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:    VSView
Short:   Refers to the vertical scrollbar used for the file navigation area.
Status:  Read/Init
Type:    OBJECTID

###DESCRIPTION###
<p>This field needs to be set to the ID of the vertical scrollbar that has
been created in the viewing area.  If you have not created a relevant vertical
scrollbar then there is no need to set this field.</p>

###SEE ALSO###
Fields: HSNavigation, HSView, VSNavigation

###END###

*****************************************************************************/

/*****************************************************************************
** Internal: FinishInstall()
*/

void FinishInstall(struct Install *Self)
{
   /*** Disable the installation buttons ***/

   if (Self->NextButtonID) acDisableID(Self->NextButtonID);
   if (Self->PrevButtonID) acDisableID(Self->PrevButtonID);
   if (Self->CancelButtonID) acDisableID(Self->CancelButtonID);

   /*** Print the completion text ***/

   if (Self->NavigationID) {
      DisplayXML(Self, "<item><b><font colour=\"#202080\">Install Complete</font></b></item><item/><item>Installation complete.  You may now quit the installer, or choose another program to install.</item>");
   }
   else DisplayXML(Self, "<item><b><font colour=\"#202080\">Install Complete</font></b></item><item/><item>Installation complete.</item>");

   /*** Reenable the navigation area ***/

   if (Self->NavigationID) acEnableID(Self->NavigationID);

   /*** Exit ***/

   if (Self->prvUninstall) {
      if (Self->prvPackageName) {
         struct mtGetXMLString getxml;
         struct acWrite write;
         UBYTE strTail[] = "</uninstall>\n";
         LONG len, i;

         DPrintF("Install:","Completing the uninstall file.");

         if (Self->prvUninstallIndex) {
            DPrintF("Install:","Writing custom uninstall tags to end of XML file.");

            getxml.Index = Self->prvUninstallIndex;
            getxml.Flags = XMF_INCLUDESIBLINGS;
            if (Action(MT_GetXMLString, Self->prvXML, &getxml) IS ERR_Okay) {
               for (i=0; getxml.Result[i]; i++);
               write.Buffer = getxml.Result;
               write.Length = i;
               Action(AC_Write, Self->prvUninstall, &write);

               write.Buffer = "\n";
               write.Length = 1;
               Action(AC_Write, Self->prvUninstall, &write);

               FreeMemory(getxml.Result);
            }
            else DPrintF("@Install:","GetXMLString() failed.");
         }

         /*** Write out the </uninstall tag at the end of the file ***/

         write.Buffer = strTail;
         write.Length = sizeof(strTail) - 1;
         Action(AC_Write, Self->prvUninstall, &write);

         /*** Move the uninstall file from its temporary location to the destination ***/

         for (len=0; Self->prvPackageName[len]; len++);

         {
            UBYTE buffer[sizeof("system:config/uninstall/")+4+len];

            MakeDirectory("system:config/uninstall", PERMIT_EVERYONE_READWRITE);

            len = StrCopy("system:config/uninstall/", buffer, COPY_ALL);
            len += StrCopy(Self->prvPackageName, buffer+len, COPY_ALL);
            len += StrCopy(".xml", buffer+len, COPY_ALL);
            ActionTags(MT_MoveFile, Self->prvUninstall, buffer);
         }
      }
      else {
         DPrintF("!Install:","Cannot produce an uninstall file as no package name was specified.");
         Action(MT_DeleteFile, Self->prvUninstall, NULL);
      }

      acFree(Self->prvUninstall);
      Self->prvUninstall = NULL;
      Self->prvUninstallIndex = 0;
   }

   /*** Refresh the desktop with a RefreshIcons call if a program icon was installed ***/

   if (Self->prvRefreshDesktop) {
      OBJECTID desktopid;
      if (FastFindObject("desktop", NULL, &desktopid, 1, NULL) IS ERR_Okay) {
         struct acCustom custom;
         custom.Number = 0;
         custom.String = "RefreshIcons";
         ActionMsg(AC_Custom, desktopid, &custom);
      }
   }

   if (Self->prvXML) {
      acFree(Self->prvXML);
      Self->prvXML = NULL;
   }

   if (Self->prvCompression) {
      acFree(Self->prvCompression);
      Self->prvCompression = NULL;
   }
}

/*****************************************************************************
** Internal: TerminateInstall()
*/

void TerminateInstall(struct Install *Self)
{
   Self->prvTag = NULL; /* No more tags to process */

   /*** Disable the installation buttons ***/

   if (Self->NextButtonID) acDisableID(Self->NextButtonID);
   if (Self->PrevButtonID) acDisableID(Self->PrevButtonID);
   if (Self->CancelButtonID) acDisableID(Self->CancelButtonID);

   /*** Reenable the navigation area ***/

   if (Self->NavigationID) acEnableID(Self->NavigationID);

   /*** Exit ***/

   if (Self->prvXML) {
      acFree(Self->prvXML);
      Self->prvXML = NULL;
      Self->prvTags = NULL;
   }

   if (Self->prvCompression) {
      acFree(Self->prvCompression);
      Self->prvCompression = NULL;
   }

   if (Self->prvUninstall) {
      Action(MT_DeleteFile, Self->prvUninstall, NULL);
      acFree(Self->prvUninstall);
      Self->prvUninstall = NULL;
   }
}

/*****************************************************************************
** Internal: CancelInstall()
*/

void CancelInstall(struct Install *Self, WORD Message)
{
   /*** Disable the installation buttons ***/

   DPrintF("CancelInstall()","[Install]");

   if (Self->Processing) {
      Self->CancelRequest = TRUE;
      return;
   }

   if (Self->NextButtonID) acDisableID(Self->NextButtonID);
   if (Self->PrevButtonID) acDisableID(Self->PrevButtonID);
   if (Self->CancelButtonID) acDisableID(Self->CancelButtonID);

   /*** Reenable the navigation area ***/

   if (Self->NavigationID) {
      if (Message) DisplayXML(Self, "<item><b><font colour=\"#202080\">Install Cancelled</font></b></item><item/><item>Installation cancelled.  You may now quit the installer, or choose another program to install.</item>");
      acEnableID(Self->NavigationID);
   }
   else if (Message) DisplayXML(Self, "<item><b><font colour=\"#202080\">Install Cancelled</font></b></item><item/><item>Installation cancelled.</item>");

   /*** Exit ***/

   if (Self->prvXML) {
      acFree(Self->prvXML);
      Self->prvXML = NULL;
   }

   if (Self->prvCompression) {
      acFree(Self->prvCompression);
      Self->prvCompression = NULL;
   }

   if (Self->prvUninstall) {
      Action(MT_DeleteFile, Self->prvUninstall, NULL);
      acFree(Self->prvUninstall);
      Self->prvUninstall = NULL;
   }
}

/*****************************************************************************
** Internal: AddVariable()
*/

static ERROR AddVariable(struct Install *Self, STRING Name, STRING Value)
{
   struct XMLTag *tag, **xmllist;
   STRING str, replace;
   LONG namelen, valuelen, pos, i, count;

   if ((!Self) OR (!Name) OR (!Value)) return(ERR_Args);

   if (Self->prvTotalVariables >= MAX_VARIABLES) return(ERR_ArrayFull);

   for (namelen=0; Name[namelen]; namelen++);
   for (valuelen=0; Value[valuelen]; valuelen++);

   /*** Check if the variable is already listed ***/

   for (pos=0; pos < Self->prvTotalVariables; pos++) {
      if (StrCompare(Name, Self->prvVariables[pos].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         FreeMemory(Self->prvVariables[pos].Name);
         Self->prvVariables[pos].Name  = NULL;
         Self->prvVariables[pos].Value = NULL;
         break;
      }
   }

   if (AllocMemory(namelen+1+valuelen+1, MEM_STRING|MEM_NOCLEAR, &str, NULL) IS ERR_Okay) {
      UBYTE variable[namelen+3];

      Self->prvVariables[pos].Name = str;
      while (*Name) *str++ = *Name++;
      *str++ = 0;

      Self->prvVariables[pos].Value = str;
      while (*Value) *str++ = *Value++;
      *str = 0;

      if (pos IS Self->prvTotalVariables) Self->prvTotalVariables++;

      /* Go through the XML tags and update anything that refers to the
      ** variable (identifiable by use of curly brackets).
      */

      variable[0] = '{';
      for (i=0; Self->prvVariables[pos].Name[i]; i++) variable[i+1] = Self->prvVariables[pos].Name[i];
      variable[i+1] = '}';
      variable[i+2] = 0;

      xmllist = Self->prvTags + Self->prvInstallIndex;
      for (count=0; count < Self->prvTotalTags; count++) {
          tag = xmllist[count];
          for (i=0; i < tag->TotalAttributes; i++) {
             if (tag->Attributes[i].Value) {
                /*** Replace all instances of the variable ***/
                if (StrReplace(tag->Attributes[i].Value, variable, Self->prvVariables[pos].Value, &replace, FALSE) IS ERR_Okay) {
                   /*** Update the XML structure ***/
                   ActionTags(MT_SetXMLAttribute, Self->prvXML, tag->Index, i, tag->Attributes[i].Name, replace);
                   tag = xmllist[count];
                   FreeMemory(replace);
                }
             }
          }
      }

      return(ERR_Okay);
   }
   else return(ERR_AllocMemory);
}

/*****************************************************************************
** Internal: CreateOption()
*/

static ERROR CreateOption(struct Install *Self, OBJECTID *ButtonID, STRING String, LONG OptionStatus)
{
   OBJECTPTR render, text, tickon, tickoff, onclick;
   LONG textheight;

   DPrintF("~CreateOption:","%s", String);

   /*** Create the option box ***/

   if (CreateObject(ID_RENDER, NF_SHARED, &render, ButtonID,
         FID_Container|TLONG, Self->ViewID,
         FID_XCoord|TLONG,    Self->prvCurrentXCoord,
         FID_YCoord|TLONG,    Self->prvCurrentYCoord,
         FID_XOffset|TLONG,   20,
         FID_Height|TLONG,    16,
         FID_Frame|TLONG,     1 + OptionStatus,
         FID_Region|TLONG,    TRUE,
         TAGEND) IS ERR_Okay) {

      /*** Create the text for the option box ***/

      CreateObject(ID_TEXT, NULL, &text, NULL,
         FID_Container|TLONG, *ButtonID,
         FID_String|TSTRING,  String,
         FID_YCoord|TLONG,    0,
         FID_Face|TSTRING,    glFontFace,
         FID_Colour|TSTRING,  "#000000",
         TAGEND);

      GetField(text, FID_TextHeight, FT_LONG, &textheight);
      SetFields(text, FID_XCoord|TLONG, textheight + 8, TAGEND);
      SetFields(render, FID_Height|TLONG, textheight, TAGEND);

      /*** Onclick frame management ***/

      CreateObject(ID_ONCLICK, NULL, &onclick, NULL,
         FID_Container|TLONG,    *ButtonID,
         FID_Flags|TLONG,        OCF_TOGGLEFRAME,
         FID_Frame|TLONG,        2 - OptionStatus,
         FID_ReleaseFrame|TLONG, 1 + OptionStatus,
         FID_Button|TLONG,       1,
         TAGEND);

      /*** Frame #1 graphics (option off) ***/

      CreateObject(ID_BOX, NULL, &tickoff, NULL,
         FID_Container|TLONG, *ButtonID,
         FID_Sunken|TLONG,    TRUE,
         FID_XCoord|TLONG,    0,
         FID_YCoord|TLONG,    0,
         FID_Width|TLONG,     textheight,
         FID_Height|TLONG,    textheight,
         FID_Colour|TSTRING,  "#f0f0f0",
         FID_Border|TSTRING,  "#000000",
         FID_Frame|TLONG,     1,
         TAGEND);

      /*** Frame #2 graphics (option on) ***/

      CreateObject(ID_BOX, NULL, &tickon, NULL,
         FID_Container|TLONG, *ButtonID,
         FID_Sunken|TLONG,    TRUE,
         FID_XCoord|TLONG,    0,
         FID_YCoord|TLONG,    0,
         FID_Width|TLONG,     textheight,
         FID_Height|TLONG,    textheight,
         FID_Colour|TSTRING,  "#00ff00",
         FID_Border|TSTRING,  "#000000",
         FID_Frame|TLONG,     2,
         TAGEND);

      Action(AC_Show, render, NULL);

      ReleaseObject(render);
   }
   else {
      StepBack();
      return(ERR_CreateObject);
   }

   Self->prvCurrentYCoord += textheight + 10;
   Self->prvIndent = textheight + 8;
   Self->prvBoxSize = textheight;

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DisplayXML()
**
** This routine should only be called when you want to create a brand new text
** page with XML content.
*/

static OBJECTPTR DisplayXML(struct Install *Self, STRING Text)
{
   OBJECTPTR text;
   LONG len;

   if ((!Self) OR (!Text)) return(NULL);

   KillChildren(Self);

   ResetPage(Self);

   if (CreateObject(ID_TEXT, NULL, &text, NULL,
      FID_Container|TLONG,    Self->ViewID,
      FID_LeftMargin|TLONG,   3,
      FID_RightMargin|TLONG,  3,
      FID_TopMargin|TLONG,    3,
      FID_BottomMargin|TLONG, 3,
      FID_Face|TSTRING,       glFontFace,
      FID_Flags|TLONG,        TXF_WORDWRAP,
      FID_VScroll|TLONG,      Self->VSViewID,
      TAGEND) IS ERR_Okay) {

      for (len=0; Text[len]; len++);

      acDataChannel(text, Self->Head.UniqueID, DATA_XML, 1, Text, len, 1);

      return(text);
   }
   else return(NULL);
}

/*****************************************************************************
** Internal: FindTag()
*/

static struct XMLTag * FindTag(struct XMLTag *XMLTag, STRING Name)
{
   struct XMLTag *tag;

   while (XMLTag) {
      if (StrCompare(Name, XMLTag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) return(XMLTag);

      if (XMLTag->Child) {
         if ((tag = FindTag(XMLTag->Child, Name))) return(tag);
      }

      XMLTag = XMLTag->Next;
   }
   return(NULL);
}

/*****************************************************************************
** Internal: KillChildren()
*/

static void KillChildren(struct Install *Self)
{
   LONG count, i;

   if ((count = TotalChildren(Self->ViewID)) > 0) {
      struct ChildEntry list[count];

      DPrintF("~KillChildren()","Destroying children in the view area.");

      if (ListChildren(Self->ViewID, list, &count) IS ERR_Okay) {
         for (i=0; i < count; i++) ActionMsg(AC_Free, list[i].ObjectID, NULL);
      }
      if (Self->ViewID) ActionMsg(AC_Draw, Self->ViewID, NULL);

      StepBack();
   }

   Self->prvOutput = NULL;
}

/*****************************************************************************
** Internal: PreviousTag()
*/

static struct XMLTag * PreviousTag(struct XMLTag *XMLTag)
{
   struct XMLTag *tag;
   LONG i;

   if (!XMLTag) return(NULL);

   for (tag=XMLTag->Prev; tag; tag=tag->Prev) {
      for (i=0; glTags[i].Tag; i++) {
         if (StrCompare(glTags[i].Tag, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) return(tag);
      }
   }

   return(XMLTag);
}

/*****************************************************************************
** Internal: SetPage()
**
** Note: Setting the page size will cause a an automatic redraw for you.
*/

static void SetPage(struct Install *Self, LONG Height)
{
   struct acRedimension resize;
   struct mtUpdateScroll update;
   OBJECTPTR scroll;

   /*** Change the height of the document page ***/

   resize.XCoord = 0;
   resize.YCoord = 0;
   resize.ZCoord = 0;
   resize.Width  = 0;
   resize.Height = Height;
   resize.Depth  = 0;
   ActionMsg(AC_Redimension, Self->ViewID, &resize);

   /*** Redirect the scroller so that it scrolls the container of the page ***/

   if (AccessObject(Self->VSViewID, 5000, &scroll) IS ERR_Okay) {
      SetFields(scroll, FID_Object|TLONG, GetOwnerID(Self->ViewID),
                        TAGEND);

      update.ViewSize = -1;
      update.PageSize = Height;
      update.Position = 0;
      update.Unit     = -1;
      Action(MT_UpdateScroll, scroll, &update);

      ReleaseObject(scroll);
   }
}

/*****************************************************************************
** Internal: ResetPage()
*/

static void ResetPage(struct Install *Self)
{
   OBJECTPTR view;
   OBJECTPTR scroll;

   if (AccessObject(Self->VSViewID, 5000, &scroll) IS ERR_Okay) {
      SetFields(scroll, FID_Object|TLONG, 0,
                        TAGEND);
      ReleaseObject(scroll);
   }

   /*** Reset the height of the document page ***/

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      SetFields(view, FID_YCoord|TLONG, 0,
                      FID_Height|TLONG, 4000,
                      TAGEND);
      ReleaseObject(view);
   }
}

/*****************************************************************************
** Internal: ScrollView()
** Short:    Calling this function regularly will keep the most recent install
**           messages in view.
*/

static void ScrollView(struct Install *Self)
{
   OBJECTPTR drawable;
   LARGE time;

   time = SystemTime();
   if ((time - Self->prvSysTime) > 500) {
      if (AccessObject(Self->ViewID, 4000, &drawable) IS ERR_Okay) {
         ActionTags(AC_Scroll, drawable, 0.0, 1000000.0, 0.0);
         ReleaseObject(drawable);
      }
      Self->prvSysTime = time;
   }
}

/*****************************************************************************
** Internal: ExecuteInstallTags()
*/

static ERROR ExecuteInstallTags(struct Install *Self, OBJECTPTR Output, struct XMLTag *XMLTag)
{
   struct mtGetXMLString getxml;
   OBJECTPTR script;
   struct XMLTag *tag;
   WORD i;
   ERROR error;

   if (!XMLTag) return(ERR_Okay); /* If there is nothing to process, everything is OK */
   if (Self->CancelRequest) return(ERR_Okay);

   error = ERR_Okay;
   for (tag=XMLTag; (tag) AND (error IS ERR_Okay); tag=tag->Next) {
      if (Self->CancelRequest) break;

      ProcessMessages(NULL, NULL);

      for (i=0; glInstallCommands[i].Tag; i++) {
         if (StrCompare(glInstallCommands[i].Tag, tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            error = glInstallCommands[i].Routine(Self, Output, tag);
            if (error != ERR_Okay) DPrintF("ExeInstallTags:","Error %d returned by tag <%s>", error, glInstallCommands[i].Tag);
            break;
         }
      }

      if (!glInstallCommands[i].Tag) {
         /* The command was not found - execute as a DML instruction.  This
         ** involves converting the tag into an XML string and passing it to
         ** the script object for execution.
         **
         ** NOTE: Child objects embedded under the tag will also be processed
         ** as part of this routine.
         */

         DPrintF("~Installer:","Processing custom tag <%s>", tag->Attributes[0].Name);

         getxml.Index = tag->Index;
         getxml.Flags = NULL;
         if (Action(MT_GetXMLString, Self->prvXML, &getxml) IS ERR_Okay) {
            if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
               FID_Statement|TSTRING, getxml.Result,
               TAGEND) IS ERR_Okay) {
               if ((error = acActivate(script)) != ERR_Okay) {
                  DPrintF("ExeInstallTags:","Error %d returned by custom tag <%s>", error, tag->Attributes[0].Name);
               }
               acFree(script);
            }
            else error = ERR_CreateObject;

            FreeMemory(getxml.Result);
         }
         else error = ERR_Failed;

         StepBack();
      }

      ScrollView(Self);

      if (Self->NoFail) error = ERR_Okay; /* If failures are going to be ignored, maintain an ignorant error code */
   }

   if (error) TerminateInstall(Self);

   return(error);
}

/****************************************************************************/

#include "commands.c"
#include "tags.c"
